# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class BaseComparator(ABC):
    SUPPORTED_EXTENSIONS = []
    
    def __init__(self, out_db_conn, excel_writer):
        self.out_db_conn = out_db_conn
        self.excel_writer = excel_writer
    
    @classmethod
    def supports(cls, file_extension):
        return file_extension in cls.SUPPORTED_EXTENSIONS
    
    @abstractmethod
    def process(self, file_a, file_b):
        pass
